# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging

import seaborn as sns
from deeponto.onto import Taxonomy
from geoopt.manifolds import PoincareBall
from sentence_transformers import SentenceTransformer

logger = logging.getLogger(__name__)


def entity_norm_plot(hierarchy: Taxonomy, model: SentenceTransformer):
    entity_names = [hierarchy.get_node_attributes(e)["name"] for e in hierarchy.nodes]
    entity_embeds = model.encode(entity_names, 1024, True, convert_to_tensor=True)
    manifold = PoincareBall(c=1 / model._first_module().get_word_embedding_dimension())
    entity_norms = manifold.dist0(entity_embeds)
    return (
        entity_embeds,
        entity_norms,
        sns.histplot(entity_norms.cpu().numpy(), bins=10, kde=True, kde_kws={"bw_adjust": 2}),
    )


def entity_depths_plot(hierarchy: Taxonomy):
    if not hierarchy.root_node:
        logger.info("No root node detected; adding in edges from current top nodes to a pseudo root node.")
        top_nodes = []
        for n in hierarchy.nodes:
            if not hierarchy.get_parents(n):
                top_nodes.append(n)
        root = "owl:Thing"
        rooted_hierarchy = Taxonomy(hierarchy.edges + [(root, t) for t in top_nodes], root_node=root)
    else:
        rooted_hierarchy = hierarchy
    depths = []
    for n in hierarchy.nodes:
        depths.append(rooted_hierarchy.get_shortest_node_depth(n))
    return depths, sns.histplot(depths, bins=10, kde=True, kde_kws={"bw_adjust": 2})



# Copyright 2024 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import random
import tempfile

import pytest
from datasets import load_dataset
from sentence_transformers.training_args import SentenceTransformerTrainingArguments

from hierarchy_transformers.losses import HierarchyTransformerLoss
from hierarchy_transformers.models import HierarchyTransformer, HierarchyTransformerTrainer


@pytest.fixture
def model_path():
    return "sentence-transformers/all-MiniLM-L6-v2"


@pytest.fixture
def dataset_path():
    return "Hierarchy-Transformers/WordNetNoun"


def test_training(model_path, dataset_path):
    # Create a temporary directory for the output files
    with tempfile.TemporaryDirectory() as temp_dir:
        # 1. Load dataset and model
        hop_type = random.choice(["MultiHop", "MixedHop"])
        neg_type = random.choice(["HardNegatives", "RandomNegatives"])
        triplet_dataset = load_dataset(dataset_path, f"{hop_type}-{neg_type}-Triplets")
        trial_train = triplet_dataset["train"].select(range(64))
        trial_val = triplet_dataset["val"].select(range(32))
        model = HierarchyTransformer.from_pretrained(model_path)

        # 2. set up the loss function
        hit_loss = HierarchyTransformerLoss(model=model)

        # 3. Define the training arguments
        args = SentenceTransformerTrainingArguments(
            output_dir=temp_dir,
            num_train_epochs=1,
            learning_rate=1e-5,
            per_device_train_batch_size=16,
            per_device_eval_batch_size=16,
            warmup_ratio=0.05,  # alternatively, set warmup_steps to 500
            eval_strategy="epoch",
            save_strategy="epoch",
        )

        # 4. Train the model on trial samples
        trainer = HierarchyTransformerTrainer(
            model=model,
            args=args,
            train_dataset=trial_train,  # train loss requires triplets
            eval_dataset=trial_val,  # val loss requires triplets
            loss=hit_loss,
        )
        trainer.train()



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import torch
import torch.nn.functional as F
from geoopt.manifolds import PoincareBall

from hierarchy_transformers.utils import format_citation


class HyperbolicEntailmentConeLoss(torch.nn.Module):
    """Hyperbolic loss that construct entailment cones for entities.

    Essentially, this loss is expected to achieve:
    $$
        angle(child, parent_cone_axis) < angle(parent_cone)
    $$

    Inputs are labelled pairs presented in `(rep_anchor, rep_other, label)`.
    """

    def __init__(self, manifold: PoincareBall, min_euclidean_norm: float = 0.1, margin: float = 0.1, eps: float = 1e-5):
        super().__init__()
        self.manifold = manifold
        assert self.manifold.c == 1.0, f"Entailment cone loss is not defined for curvature: {manifold.c}."
        self.min_euclidean_norm = min_euclidean_norm
        self.margin = margin
        self.eps = eps

    def get_config_dict(self):
        config = {"distance_metric": "PoincareBall(c=1.0).cone_angle", "margin": self.margin}
        return config

    def half_cone_aperture(self, cone_tip: torch.Tensor):
        """Angle between the axis [0, x] (line through 0 and x) and the boundary of the cone at x,
        where x is the cone tip.
        """
        # cone tip means the point x is the tip of the hyperbolic cone
        # norm_tip = cone_tip.norm(dim=-1).clamp(min=self.min_euclidean_norm)  # to prevent undefined aperture
        sq_norm_tip = cone_tip.pow(2).sum(dim=-1).clamp(min=self.min_euclidean_norm + self.eps, max=1 - self.eps)
        return torch.arcsin(self.min_euclidean_norm * (1 - sq_norm_tip) / torch.sqrt(sq_norm_tip)).clamp(
            min=-1 + self.eps, max=1 - self.eps
        )

    def cone_angle_at_u(self, cone_tip: torch.Tensor, u: torch.Tensor):
        """Angle between the axis [0, x] and the line [x, u]. This angle should be smaller than the
        half cone aperture at x for real children.
        """
        # parent point is treated as the cone tip
        norm_tip = cone_tip.norm(2, dim=-1)
        norm_child = u.norm(2, dim=-1)
        dot_prod = (cone_tip * u).sum(dim=-1)
        edist = (cone_tip - u).norm(2, dim=-1)  # euclidean distance
        numerator = dot_prod * (1 + norm_tip**2) - norm_tip**2 * (1 + norm_child**2)
        denominator = norm_tip * edist * torch.sqrt(1 + (norm_child**2) * (norm_tip**2) - 2 * dot_prod)

        angle = torch.arccos((numerator / denominator.clamp(min=self.eps)).clamp(min=-1 + self.eps, max=1 - self.eps))
        # Debugging step
        if torch.isnan(angle).any():
            print("Numerator:", numerator)
            print("Denominator:", denominator)
            print("Angle calculation resulted in NaNs")

        return angle

    def energy(self, cone_tip: torch.Tensor, u: torch.Tensor):
        """Enery function defined as: max(0, cone_angle(u) - half_cone_aperture) given a cone tip."""
        return F.relu(self.cone_angle_at_u(cone_tip, u) - self.half_cone_aperture(cone_tip))

    def forward(self, rep_anchor: torch.Tensor, rep_other: torch.Tensor, labels: torch.Tensor):
        # anchors are children
        energies = self.energy(cone_tip=rep_other, u=rep_anchor)
        cone_loss = labels.float() * energies + (1 - labels).float() * F.relu(self.margin - energies)
        return cone_loss.mean()

    @property
    def citation(self) -> str:
        return format_citation(
            """
            @inproceedings{ganea2018hyperbolic,
              title={Hyperbolic entailment cones for learning hierarchical embeddings},
              author={Ganea, Octavian and B{\'e}cigneul, Gary and Hofmann, Thomas},
              booktitle={International conference on machine learning},
              pages={1646--1655},
              year={2018},
              organization={PMLR}
            }
            """
        )


class HyperbolicEntailmentConeTripletLoss(HyperbolicEntailmentConeLoss):
    """Hyperbolic loss that construct entailment cones for entities.

    Essentially, this loss is expected to achieve:
    $$
        angle(child, parent_cone_axis) < angle(parent_cone)
    $$

    Inputs are triplets presented in `(rep_anchor, rep_positive, rep_negative)`.
    """

    def __init__(self, manifold: PoincareBall, min_euclidean_norm: float = 0.1, margin: float = 0.1, eps: float = 1e-5):
        super().__init__(manifold, min_euclidean_norm, margin, eps)

    def forward(self, rep_anchor: torch.Tensor, rep_positive: torch.Tensor, rep_negative: torch.Tensor):
        # anchors are children
        energies_positive = self.energy(cone_tip=rep_positive, u=rep_anchor)
        energies_negative = self.energy(cone_tip=rep_negative, u=rep_anchor)
        cone_triplet_loss = F.relu(energies_positive - energies_negative + self.margin)
        return cone_triplet_loss.mean()


class HyperbolicEntailmentConeStaticLoss(HyperbolicEntailmentConeLoss):
    """Hyperbolic loss that construct entailment cones for entities.

    Essentially, this loss is expected to achieve:
    $$
        angle(child, parent_cone_axis) < angle(parent_cone)
    $$

    Inputs are presented in `(subject, *objects)` where the first `object` is positive and the rest are negative.

    This is designed for the static embedding implementation.
    """

    def __init__(self, manifold: PoincareBall, min_euclidean_norm: float = 0.1, margin: float = 0.1, eps: float = 1e-5):
        super().__init__(manifold, min_euclidean_norm, margin, eps)

    def forward(self, subject: torch.Tensor, objects: torch.Tensor):
        # the first object is positive
        energy = self.energy(objects, subject)
        return (energy[:, 0].sum() + F.relu(self.margin - energy[:, 1:]).sum()) / torch.numel(energy)




# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from .hit_loss import HierarchyTransformerLoss, HyperbolicCentripetalLoss, HyperbolicClusteringLoss
from .hyper_cone_loss import (
    HyperbolicEntailmentConeLoss,
    HyperbolicEntailmentConeStaticLoss,
    HyperbolicEntailmentConeTripletLoss,
)
from .poincare_embed_loss import PoincareEmbeddingStaticLoss

__all__ = [
    "HierarchyTransformerLoss",
    "HyperbolicCentripetalLoss",
    "HyperbolicClusteringLoss",
    "HyperbolicEntailmentConeLoss",
    "HyperbolicEntailmentConeStaticLoss",
    "HyperbolicEntailmentConeTripletLoss",
    "PoincareEmbeddingStaticLoss",
]



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import torch
from geoopt.manifolds import PoincareBall

from hierarchy_transformers.utils import format_citation


class PoincareEmbeddingStaticLoss(torch.nn.Module):
    """Poincare embedding loss.

    Essentially, this loss is expected to achieve:

    $$d(child, parent) < d(child, negative)$$

    Inputs are presented in `(subject, *objects)` where the first `object` is positive and the rest are negative.

    This is designed for the static embedding implementation.
    """

    def __init__(self, manifold: PoincareBall):
        super().__init__()
        self.manifold = manifold
        self.cross_entropy = torch.nn.CrossEntropyLoss()

    def forward(self, subject: torch.Tensor, objects: torch.Tensor):
        # first object is always the correct one
        pred_dists = self.manifold.dist(subject, objects)
        correct_object_indices = torch.tensor([0] * len(pred_dists)).to(pred_dists.device)
        return self.cross_entropy(-pred_dists, correct_object_indices)

    @property
    def citation(self) -> str:
        return format_citation(
            """
            @article{nickel2017poincare,
              title={Poincar{\'e} embeddings for learning hierarchical representations},
              author={Nickel, Maximillian and Kiela, Douwe},
              journal={Advances in neural information processing systems},
              volume={30},
              year={2017}
            }
            """
        )



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging

import torch
from geoopt import ManifoldParameter
from geoopt.manifolds import PoincareBall

logger = logging.getLogger(__name__)


class PoincareStaticEmbedding(torch.nn.Module):
    r"""Class for the static hyperbolic embedding models:

        - [1] Poincaré Embedding by [Nickel et al., NeurIPS 2017](https://arxiv.org/abs/1705.08039).
        - [2] Hyperbolic Entailment Cone by [Ganea et al., ICML 2018](https://arxiv.org/abs/1804.01882).

    both of which lie in a unit Poincaré ball. According to [2], it is better to apply the entailment cone loss in the post-training phase of a Poincaré embedding model in [1].

    Attributes:
        entities (list): The list of input entity IDs (fixed).
        idx2ent (dict): A dictionary that stores the `(index, entity_id)` pairs.
        ent2idx (dict): A dictionary that stores the `(entity_id, index)` pairs.
        embed_dim (int): The embedding dimension of this model.
        manifold (geoopt.manifolds.PoincareBall): The hyperbolic manifold (Poincaré Ball) of this model.
        embed (torch.nn.Embedding): The static hyperbolic embeddings for entities.
    """

    def __init__(self, entity_ids: list, embed_dim: int, init_weights: float = 1e-3):
        super().__init__()

        self.entities = entity_ids
        self.idx2ent = {idx: ent for idx, ent in enumerate(self.entities)}
        self.ent2idx = {v: k for k, v in self.idx2ent.items()}
        self.embed_dim = embed_dim
        self.manifold = PoincareBall()
        self.dist = self.manifold.dist

        # initialise static embedding
        self.embed = torch.nn.Embedding(
            num_embeddings=len(self.idx2ent),  # fixed num embeddings,
            embedding_dim=self.embed_dim,
            sparse=False,
            max_norm=1.0,  # unit poincare ball projection
        )
        self.embed.weight.data.uniform_(-init_weights, init_weights)
        self.embed.weight = ManifoldParameter(self.embed.weight, manifold=self.manifold)
        logger.info(f"Init static hyperbolic embedding for {len(self.idx2ent)} entities.")

    def forward(self, inputs: torch.Tensor):
        """Forward propagation.

        The inputs are organised as `(batch_size, num_entities, embed_dim)` where `dim=` includes `(child, parent, negative_parents*)`.
        """

        input_embeds = self.embed(
            inputs
        )  # (batch_size, num_entities, hidden_dim), dim 1 includes (child, parent, negative_parents*)
        objects = input_embeds.narrow(dim=1, start=1, length=input_embeds.size(1) - 1)  # use .narrow to keep dim
        subject = input_embeds.narrow(dim=1, start=0, length=1).expand_as(objects)
        return subject, objects



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from .hit_eval import HierarchyTransformerEvaluator
from .sbert_eval import SentenceTransformerEvaluator
from .static_embed_eval import PoincareStaticEmbeddingEvaluator

__all__ = [
    "HierarchyTransformerEvaluator",
    "SentenceTransformerEvaluator",
    "PoincareStaticEmbeddingEvaluator",
]



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import torch
from geoopt.manifolds import PoincareBall


def get_circum_poincareball(embed_dim: int) -> PoincareBall:
    """Get a Poincaré Ball with a curvature adapted to a given embedding dimension so that it circumscribes the output embedding space of pre-trained language models."""
    curvature = 1 / embed_dim
    manifold = PoincareBall(c=curvature)
    return manifold


def project_onto_subspace(manifold: PoincareBall, point: torch.Tensor, normal: torch.Tensor):
    """Compute the (hyperbolic) projection of a point onto a subspace (a hyper-plane through origin) of the input manifold.

    The projected point is the mid point of the geodesic segment that joins the input point and its reflection point about the subspace plane.

    Args:
        manifold (geoopt.manifolds.PoincareBall): The input Poincaré ball manifold.
        point (torch.Tensor): The input point
        normal (torch.Tensor): The normal vector of the subspace.
    """
    reflection = reflect_about_subspace(point, normal)
    midpoint = manifold.weighted_midpoint(torch.vstack((point, reflection)))
    return midpoint


def reflect_about_subspace(point: torch.Tensor, normal: torch.Tensor):
    """Compute the (Euclidean) reflection of a point about a sub-space (a hyper-plane through origin).

    This is a helper function for computing hyperbolic subspace projection of a point.

    Args:
        point (torch.Tensor): The input point.
        normal (torch.Tensor): The normal vector of the plane (through orgin).
    """

    # Ensure the norm vector is non-zero
    if torch.all(normal == 0):
        raise ValueError("Norm vector cannot be zero.")

    # Calculate the dot product and magnitude squared of the norm vector
    dot_product = torch.dot(point, normal)
    normal_squared = torch.dot(normal, normal)

    # Compute the reflection point without explicitly normalizing the norm vector
    reflection = point - 2 * (dot_product / normal_squared) * normal

    return reflection



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from textwrap import dedent

import torch


def get_torch_device(gpu_id: int):
    return torch.device(f"cuda:{gpu_id}" if torch.cuda.is_available() else "cpu")


def are_models_equal(model1: torch.nn.Module, model2: torch.nn.Module, tolerance: float = 1e-6) -> bool:
    """
    Compares two PyTorch models to check if they are the same by comparing each parameter and buffer.

    Args:
        model1 (torch.nn.Module): The first model to compare.
        model2 (torch.nn.Module): The second model to compare.
        tolerance (float): The tolerance level for floating-point comparison (default is 1e-6).

    Returns:
        bool: True if the models are the same, False otherwise.
    """
    # Compare model parameters
    for param1, param2 in zip(model1.parameters(), model2.parameters()):
        if not torch.allclose(param1, param2, atol=tolerance):
            return False

    # Compare model buffers (e.g., running mean/variance in batch norm layers)
    for buffer1, buffer2 in zip(model1.buffers(), model2.buffers()):
        if not torch.allclose(buffer1, buffer2, atol=tolerance):
            return False

    return True


def format_citation(bibtex: str):
    """
    Use `dedent` to properly form bibtex string.
    """
    return dedent(bibtex)



# Copyright 2024 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Any

import torch
from sentence_transformers.trainer import SentenceTransformerTrainer

from .hit import HierarchyTransformer


class HierarchyTransformerTrainer(SentenceTransformerTrainer):
    r"""A simple extension of `SentenceTransformerTrainer` to monitor and log batch losses of `HierarchyTransformer`."""

    def compute_loss(
        self,
        model: HierarchyTransformer,
        inputs: dict[str, torch.Tensor | Any],
        return_outputs: bool = False,
        num_items_in_batch=None,
    ) -> torch.Tensor | tuple[torch.Tensor, dict[str, Any]]:
        loss_dict = super().compute_loss(
            model=model, inputs=inputs, return_outputs=return_outputs, num_items_in_batch=num_items_in_batch
        )
        outputs = None
        if return_outputs:
            loss_dict, outputs = loss_dict
        self.log(
            {
                "cluster_loss": round(loss_dict["cluster_loss"].item(), 4),
                "centri_loss": round(loss_dict["centri_loss"].item(), 4),
                "combined_loss": round(loss_dict["loss"].item(), 4),
            }
        )

        return (loss_dict["loss"], outputs) if return_outputs else loss_dict["loss"]



# Copyright 2024 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import random

import pytest
from datasets import load_dataset


@pytest.fixture(params=os.getenv("DATASET_PATHS", "").split(","))
def dataset_path(request):
    # Ensure there are valid dataset names
    if not request.param:
        pytest.fail("No valid dataset names found in the DATASET_PATHS environment variable")
    return request.param.strip()  # Strip any extra spaces


def test_dataset_loading(dataset_path):
    hop_type = random.choice(["MultiHop", "MixedHop"])
    neg_type = random.choice(["HardNegatives", "RandomNegatives"])
    struct_type = random.choice(["Triplets", "Pairs"])
    part = f"{hop_type}-{neg_type}-{struct_type}"
    try:
        # Attempt to load the HierarchyTransformer model
        dataset = load_dataset(dataset_path, part)
    except Exception as e:
        pytest.fail(f"Dataset failed to load: {str(e)}")

    # Check that the datast is not None
    assert dataset is not None, "Loaded dataset is None"


# [Deprecated] HF upload code

# save_path = "WordNetNoun"
# hop_type = "MultiHop"
# for split in ["train", "val", "test"]:
#     n_rand = 0
#     n_hard = 0
#     triplet_examples = {
#         "random_negatives": [],
#         "hard_negatives": []
#     }
#     pair_examples = {
#         "random_negatives": [],
#         "hard_negatives": []
#     }
#     for sample in dataset[split]:
#         child = entity_lexicon[sample["child"]]["name"]
#         parent = entity_lexicon[sample["parent"]]["name"]
#         negative_parents = [entity_lexicon[neg]["name"] for neg in sample["random_negatives"]]
#         n_rand += len(negative_parents)
#         hard_negatives = [entity_lexicon[sib]["name"] for sib in sample["hard_negatives"]]
#         n_hard += len(hard_negatives)
#         triplet_examples["random_negatives"] += [(child, parent, neg) for neg in negative_parents]
#         triplet_examples["hard_negatives"] += [(child, parent, neg) for neg in hard_negatives]
#         pair_examples["random_negatives"] += [(child, parent, 1)] + [(child, neg, 0) for neg in negative_parents]
#         pair_examples["hard_negatives"] += [(child, parent, 1)] + [(child, neg, 0) for neg in hard_negatives]
#     assert n_rand == n_hard
#     pd.DataFrame(triplet_examples["random_negatives"], columns=["child", "parent", "negative"]).to_parquet(f"{save_path}/{hop_type}-RandomNegatives-Triplets/{split}.parquet", index=False)
#     pd.DataFrame(triplet_examples["hard_negatives"], columns=["child", "parent", "negative"]).to_parquet(f"{save_path}/{hop_type}-HardNegatives-Triplets/{split}.parquet", index=False)
#     pd.DataFrame(pair_examples["random_negatives"], columns=["child", "parent", "label"]).to_parquet(f"{save_path}/{hop_type}-RandomNegatives-Pairs/{split}.parquet", index=False)
#     pd.DataFrame(pair_examples["hard_negatives"], columns=["child", "parent", "label"]).to_parquet(f"{save_path}/{hop_type}-HardNegatives-Pairs/{split}.parquet", index=False)

# [Deprecated] Compare HF and local version
# data_path = "/home/yuan/projects/HiT/data/wordnet-multi"
# dataset, entity_lexicon = load_hierarchy_dataset(data_path)

# for split in ["train", "val", "test"]:
#     for is_hard in [True, False]:
#         for is_triplet in [True, False]:
#             dataset_zenodo = prepare_hierarchy_examples(entity_lexicon, dataset[split], is_hard, is_triplet)
#             if is_triplet:
#                 dataset_zenodo = [{'child': x.texts[0], 'parent': x.texts[1], 'negative': x.texts[2]} for x in dataset_zenodo]
#             else:
#                 dataset_zenodo = [{'child': x.texts[0], 'parent': x.texts[1], 'label': x.label} for x in dataset_zenodo]
#             neg = "HardNegatives" if is_hard else "RandomNegatives"
#             struct = "Triplets" if is_triplet else "Pairs"
#             dataset_hf = load_dataset("Hierarchy-Transformers/WordNetNoun", f"MultiHop-{neg}-{struct}")[split]
#             for i in tqdm(range(len(dataset_zenodo)), desc=f"Check MultiHop-{neg}-{struct}"):
#                 assert dataset_zenodo[i] == dataset_hf[i], (dataset_zenodo[i], dataset_hf[i])



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
from collections.abc import Iterable

import torch
import torch.nn.functional as F
from geoopt.manifolds import PoincareBall

from hierarchy_transformers.models import HierarchyTransformer
from hierarchy_transformers.utils import format_citation

logger = logging.getLogger(__name__)


class HierarchyTransformerLoss(torch.nn.Module):
    """Hyperbolic loss that linearly combines hperbolic clustering loss and hyperbolic Centripetal loss and applies weights for joint optimisation."""

    def __init__(
        self,
        model: HierarchyTransformer,
        clustering_loss_weight: float = 1.0,
        clustering_loss_margin: float = 5.0,
        centripetal_loss_weight: float = 1.0,
        centripetal_loss_margin: float = 0.5,
    ):
        super().__init__()

        self.model = model
        self.manifold = self.model.manifold
        self.cluster_loss = HyperbolicClusteringLoss(self.model.manifold, clustering_loss_margin)
        self.centri_loss = HyperbolicCentripetalLoss(self.model.manifold, centripetal_loss_margin)
        self.cluster_weight = clustering_loss_weight
        self.centri_weight = centripetal_loss_weight

    def get_config_dict(self):
        # distance_metric_name = self.distance_metric.__name__
        config = {"distance_metric": f"PoincareBall(c={self.manifold.c}).dist and dist0"}
        config[HyperbolicClusteringLoss.__name__] = {
            "weight": self.cluster_weight,
            **self.cluster_loss.get_config_dict(),
        }
        config[HyperbolicCentripetalLoss.__name__] = {
            "weight": self.centri_weight,
            **self.centri_loss.get_config_dict(),
        }
        return config

    def forward(self, sentence_features: Iterable[dict[str, torch.Tensor]], labels: torch.Tensor):
        """Forward propagation that follows [`sentence_transformers.losses`](https://github.com/UKPLab/sentence-transformers/tree/master/sentence_transformers/losses)."""
        reps = [self.model(sentence_feature)["sentence_embedding"] for sentence_feature in sentence_features]
        assert len(reps) == 3
        rep_anchor, rep_positive, rep_negative = reps

        # compute and combine hyperbolic clustering and centripetal losses
        cluster_loss = self.cluster_loss(rep_anchor, rep_positive, rep_negative)
        centri_loss = self.centri_loss(rep_anchor, rep_positive, rep_negative)
        combined_loss = self.cluster_weight * cluster_loss + self.centri_weight * centri_loss

        return {
            "loss": combined_loss,
            "cluster_loss": cluster_loss,
            "centri_loss": centri_loss,
        }

    @property
    def citation(self) -> str:
        return format_citation(
            """
            @article{he2024language,
              title={Language models as hierarchy encoders},
              author={He, Yuan and Yuan, Zhangdie and Chen, Jiaoyan and Horrocks, Ian},
              journal={arXiv preprint arXiv:2401.11374},
              year={2024}
            }
            """
        )


class HyperbolicClusteringLoss(torch.nn.Module):
    r"""Hyperbolic loss that clusters entities in subsumptions.

    Essentially, this loss is expected to achieve:

    $$d(child, parent) < d(child, negative)$$

    Inputs are presented in `(rep_anchor, rep_positive, rep_negative)`.
    """

    def __init__(self, manifold: PoincareBall, margin: float):
        super().__init__()
        self.manifold = manifold
        self.margin = margin

    def get_config_dict(self):
        config = {
            "distance_metric": f"PoincareBall(c={self.manifold.c}).dist",
            "margin": self.margin,
        }
        return config

    def forward(self, rep_anchor: torch.Tensor, rep_positive: torch.Tensor, rep_negative: torch.Tensor):
        """Forward propagation.

        Args:
            rep_anchor (torch.Tensor): The input tensor for child entities.
            rep_positive (torch.Tensor): The input tensor for parent entities.
            rep_negative (torch.Tensor): The input tensor for negative parent entities.
        """
        distances_positive = self.manifold.dist(rep_anchor, rep_positive)
        distances_negative = self.manifold.dist(rep_anchor, rep_negative)
        cluster_triplet_loss = F.relu(distances_positive - distances_negative + self.margin)
        return cluster_triplet_loss.mean()

    @property
    def citation(self) -> str:
        return format_citation(
            """
            @article{he2024language,
              title={Language models as hierarchy encoders},
              author={He, Yuan and Yuan, Zhangdie and Chen, Jiaoyan and Horrocks, Ian},
              journal={arXiv preprint arXiv:2401.11374},
              year={2024}
            }
            """
        )


class HyperbolicCentripetalLoss(torch.nn.Module):
    r"""Hyperbolic loss that regulates the norms of child and parent entities.

    Essentially, this loss is expected to achieve:

    $$d(child, origin) > d(parent, origin)$$

    Inputs are presented in `(rep_anchor, rep_positive, rep_negative)` but only `(rep_anchor, rep_positive)` pairs are involved in this loss.
    """

    def __init__(self, manifold: PoincareBall, margin: float):
        super().__init__()
        self.manifold = manifold
        self.margin = margin

    def get_config_dict(self):
        config = {
            "distance_metric": f"PoincareBall(c={self.manifold.c}).dist0",
            "margin": self.margin,
        }
        return config

    def forward(self, rep_anchor: torch.Tensor, rep_positive: torch.Tensor, rep_negative: torch.Tensor):
        """Forward propagation.

        Args:
            rep_anchor (torch.Tensor): The input tensor for child entities.
            rep_positive (torch.Tensor): The input tensor for parent entities.
            rep_negative (torch.Tensor): The input tensor for negative parent entities (actually not required in this loss).
        """
        rep_anchor_hyper_norms = self.manifold.dist0(rep_anchor)
        rep_positive_hyper_norms = self.manifold.dist0(rep_positive)
        # child further than parent w.r.t. origin
        centri_triplet_loss = F.relu(self.margin + rep_positive_hyper_norms - rep_anchor_hyper_norms)
        return centri_triplet_loss.mean()

    @property
    def citation(self) -> str:
        return format_citation(
            """
            @article{he2024language,
              title={Language models as hierarchy encoders},
              author={He, Yuan and Yuan, Zhangdie and Chen, Jiaoyan and Horrocks, Ian},
              journal={arXiv preprint arXiv:2401.11374},
              year={2024}
            }
            """
        )



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import json
from pathlib import Path

from deeponto.onto import Taxonomy, TaxonomyNegativeSampler
from deeponto.utils import save_file
from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm


class HierarchyDatasetConstructor:
    """
    Class for contructing hierarchy understanding datasets.
    """
    def __init__(self, hierarchy: Taxonomy):
        self.hierarchy = hierarchy
        self.neg_sampler = TaxonomyNegativeSampler(self.hierarchy)

    def get_hard_negative(self, entity_id: str):
        """
        Get a hard negative subsumer (sibling) for the input entity.
        """
        parents = self.hierarchy.get_parents(entity_id)
        ancestors = self.hierarchy.get_parents(entity_id, True)
        siblings = []
        for parent in parents:
            siblings += self.hierarchy.get_children(parent)
        hard_negatives = set(siblings) - set([entity_id]) - set(ancestors)
        return list(hard_negatives)

    def get_transitive_edges(self, edges: list):
        """
        Get all indirect subsumptions by transitive reasoning.
        """
        trans_edges = []
        for child, _ in edges:
            trans_edges += [(child, parent) for parent in self.hierarchy.get_parents(child, True)]
        return list(set(trans_edges) - set(edges))

    def save_entity_lexicon(self, output_dir: str):
        """
        Save the entity lexicon.
        """
        entity_lexicon = dict()
        for n in self.hierarchy.nodes:
            entity_lexicon[n] = self.hierarchy.get_node_attributes(n)
        save_file(entity_lexicon, f"{output_dir}/entity_lexicon.json")

    def save_dataset(self, dataset: list, output_file: str):
        """
        Save the constructed dataset.
        """
        with open(f"{output_file}", "w+") as f:
            f.writelines("\n".join([json.dumps(sample) for sample in dataset]))

    def construct_example(self, child: str, parent: str, num_negative: int = 10):
        """
        Construct negative examples given a positive `(child, parent)` pair.
        """
        example = {"child": child, "parent": parent}
        example["random_negatives"] = self.neg_sampler.sample(child, num_negative)
        example["hard_negatives"] = (self.get_hard_negative(child) + example["random_negatives"])[:num_negative]
        return example

    def construct(self, output_dir: str, num_negative: int = 10, eval_size=0.1):
        """
        Construct the multi-hop and mixed-hop datasets.
        """
        Path(output_dir).mkdir(parents=True, exist_ok=True)

        base_edges = [(child, parent) for parent, child in self.hierarchy.edges]
        trans_edges = self.get_transitive_edges(base_edges)
        assert not set(trans_edges).intersection(set(base_edges))

        base_examples = []
        for child, parent in tqdm(base_edges, desc="base"):
            base_examples.append(self.construct_example(child, parent, num_negative))

        trans_examples = []
        for child, parent in tqdm(trans_edges, desc="trans"):
            trans_examples.append(self.construct_example(child, parent, num_negative))

        _, trans_eval_examples = train_test_split(trans_examples, test_size=eval_size)
        trans_val_examples, trans_test_examples = train_test_split(trans_eval_examples, test_size=0.5)

        trans_task_name = "multi"
        Path(f"{output_dir}/{trans_task_name}").mkdir(parents=True, exist_ok=True)
        self.save_dataset(base_examples, f"{output_dir}/{trans_task_name}/train.jsonl")
        # self.save_dataset(trans_train_examples, f"{output_dir}/transitivity/trans_train.jsonl")
        self.save_dataset(trans_val_examples, f"{output_dir}/{trans_task_name}/val.jsonl")
        self.save_dataset(trans_test_examples, f"{output_dir}/{trans_task_name}/test.jsonl")
        self.save_entity_lexicon(f"{output_dir}/{trans_task_name}")

        base_train_examples, base_eval_examples = train_test_split(base_examples, test_size=eval_size)
        base_val_examples, base_test_examples = train_test_split(base_eval_examples, test_size=0.5)
        # base_train_edges = [(x["child"], x["parent"]) for x in base_train_examples]
        # trans_base_train_edges = self.get_transitive_edges(base_train_edges)
        # trans_base_train_examples = []
        # for child, parent in tqdm(trans_base_train_edges, desc="trans on base_train"):
        #     trans_base_train_examples.append(self.construct_example(child, parent, num_negative))
        pred_task_name = "mixed"
        Path(f"{output_dir}/{pred_task_name}").mkdir(parents=True, exist_ok=True)
        self.save_dataset(base_train_examples, f"{output_dir}/{pred_task_name}/train.jsonl")
        # self.save_dataset(trans_base_train_examples, f"{output_dir}/induc/trans_base_train.jsonl")
        self.save_dataset(base_val_examples + trans_val_examples, f"{output_dir}/{pred_task_name}/val.jsonl")
        self.save_dataset(base_test_examples + trans_test_examples, f"{output_dir}/{pred_task_name}/test.jsonl")
        self.save_entity_lexicon(f"{output_dir}/{pred_task_name}")



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import math

import torch
from tqdm import tqdm


def f1_score(predictions: torch.Tensor, labels: torch.Tensor, truth_label: int = 1):
    """Pytorch tensor implementation of `f1_score` computation.

    Args:
        predictions (torch.Tensor): Predictions.
        labels (torch.Tensor): Reference labels.
        truth_label (int, optional): Specify which label represents the truth. Defaults to `1`.

    Returns:
        results (dict): result dictionary containing `Precision`, `Recall`, and `F1`.
    """
    tp = torch.sum((labels == truth_label) & (predictions == truth_label))  # correct and positive
    fp = torch.sum((labels != truth_label) & (predictions == truth_label))  # incorrect but positive
    fn = torch.sum((labels == truth_label) & (predictions != truth_label))  # correct but negative
    precision = tp / (tp + fp)
    recall = tp / (tp + fn)
    f1 = 2 * (precision * recall) / (precision + recall)
    return {"precision": precision.item(), "recall": recall.item(), "f1": f1.item()}  # recall is ACC+


def accurarcy(predictions: torch.Tensor, labels: torch.Tensor):
    """Pytorch tensor implementation of `accuracy` computation."""
    acc = torch.sum(labels == predictions) / len(labels)
    return {"accuracy": acc.item()}


def accurarcy_on_negatives(predictions: torch.Tensor, labels: torch.Tensor, truth_label: int = 1):
    """Pytorch tensor implementation of `accuracy_on_negatives` computation.

    That is, it computes accuracy only w.r.t. negative samples (with `label != truth_label`).
    """
    neg_acc = torch.sum((labels == predictions) & (labels != truth_label)) / torch.sum(labels != truth_label)
    return {"accuracy_on_negatives": neg_acc.item()}


def evaluate_by_threshold(
    scores: torch.Tensor,
    labels: torch.Tensor,
    threshold: float,
    truth_label: int = 1,
    smaller_scores_better: bool = False,
):
    r"""Compute evaluation metrics (`Precision`, `Recall`, `F1`, `Accurarcy`, `Accurarcy-`) based on the threshold.

    Args:
        scores (torch.Tensor): Prediction scores.
        labels (torch.Tensor): Reference labels.
        threshold (float): Threshold that splits the positive and negative predictions.
        truth_label (int): Specify which label represents the truth. Defaults to `1`.
        smaller_scores_better (bool): Specify if smaller than threshold indicates positive or not. Defaults to `False`.
    """

    # thresholding
    if smaller_scores_better:
        predictions = scores <= threshold
    else:
        predictions = scores > threshold
    # compute results
    results = {
        "threshold": threshold,
        **f1_score(predictions=predictions, labels=labels, truth_label=truth_label),
        **accurarcy(predictions=predictions, labels=labels),
        **accurarcy_on_negatives(predictions=predictions, labels=labels, truth_label=truth_label),
    }
    return results


def grid_search(
    scores: torch.Tensor,
    labels: torch.Tensor,
    threshold_granularity: int = 100,
    truth_label: int = 1,
    smaller_scores_better: bool = False,
    primary_metric: str = "f1",
    best_primary_metric_value: float = -math.inf,
    preformatted_best_results: dict = {},
):
    """Grid search the best scoring threshold.

    Args:
        scores (torch.Tensor): Prediction scores.
        labels (torch.Tensor): Reference labels.
        threshold_granularity (int, optional): A score scaling factor to determine the granularity of grid search. Defaults to `100`.
        truth_label (int): Specify which label represents the truth. Defaults to `1`.
        smaller_scores_better (bool): Specify if smaller than threshold indicates positive or not. Defaults to `False`.
        primary_metric (str, optional): The primary evaluation metric to determine the grid search result. Defaults to `"F1"`.
        best_primary_metric_value (Optional[float], optional): Best previous primary metric value. Defaults to `-math.inf`.
        preformatted_best_results (dict, optional): Preformatted best results dictionary. Defaults to `{}`.
    """

    best_results = None

    # grid search start and end are confined by the prediction scores
    start = int(scores.min() * threshold_granularity)
    end = int(scores.max() * threshold_granularity)

    # grid search to update the best results
    for threshold in tqdm(range(start, end), desc="Thresholding"):
        threshold = threshold / threshold_granularity
        results = evaluate_by_threshold(
            scores=scores,
            labels=labels,
            threshold=threshold,
            truth_label=truth_label,
            smaller_scores_better=smaller_scores_better,
        )
        if results[primary_metric] >= best_primary_metric_value:
            best_results = preformatted_best_results
            best_results.update(results)
            best_primary_metric_value = results[primary_metric]

    return best_results



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging

import torch
from geoopt.optim import RiemannianAdam
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import get_linear_schedule_with_warmup

from .poincare_embed import PoincareStaticEmbedding

logger = logging.getLogger(__name__)


class PoincareStaticEmbeddingTrainer:
    r"""Class for training the static hyperbolic embedding models:

        - [1] Poincaré Embedding by [Nickel et al., NeurIPS 2017](https://arxiv.org/abs/1705.08039).
        - [2] Hyperbolic Entailment Cone by [Ganea et al., ICML 2018](https://arxiv.org/abs/1804.01882).

    both of which lie in a unit Poincaré ball. According to [2], it is better to apply the entailment cone loss in the post-training phase of a Poincaré embedding model in [1].
    """

    def __init__(
        self,
        model: PoincareStaticEmbedding,
        train_dataset: list,
        loss: torch.nn.Module,
        num_train_epochs: int = 256,
        learning_rate: float = 0.01,
        train_batch_size: int = 200,
        warmup_epochs: int = 10,
    ):
        self.model = model
        self.train_dataloader = DataLoader(torch.tensor(train_dataset), shuffle=True, batch_size=train_batch_size)
        self.loss = loss
        self.learning_rate = learning_rate
        self.optimizer = RiemannianAdam(self.model.parameters(), lr=self.learning_rate)
        self.current_epoch = 0
        self.num_train_epochs = num_train_epochs
        self.num_epoch_steps = len(self.train_dataloader)
        self.num_training_steps = self.num_epoch_steps * self.num_train_epochs
        self.warmup_epochs = warmup_epochs
        self.scheduler = get_linear_schedule_with_warmup(
            self.optimizer,
            num_warmup_steps=self.warmup_epochs * self.num_epoch_steps,  # one epoch warming-up
            num_training_steps=self.num_training_steps,
        )

    @property
    def lr(self):
        for g in self.optimizer.param_groups:
            return g["lr"]

    def training_step(self, batch, device):
        batch = batch.to(device)
        self.optimizer.zero_grad(set_to_none=True)
        subject, objects = self.model(batch)
        loss = self.loss(subject, objects)
        loss.backward()
        self.optimizer.step()
        self.scheduler.step()
        return loss

    def train(self, device):
        self.model.to(device)
        for _ in range(self.num_train_epochs):
            epoch_bar = tqdm(
                range(self.num_epoch_steps), desc=f"Epoch {self.current_epoch + 1}", leave=True, unit="batch"
            )
            for batch in self.train_dataloader:
                loss = self.training_step(batch, device)
                epoch_bar.set_postfix({"batch_loss": loss.item(), "lr": self.lr})
                epoch_bar.update()
            self.current_epoch += 1



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import os

from datasets import Dataset, load_dataset
from tqdm import tqdm

logger = logging.getLogger(__name__)


def load_hf_dataset(path: str, name: str | None = None, **config_kwargs):
    """Load a HiT dataset from Hugging Face.

    See available datasets on: https://huggingface.co/Hierarchy-Transformers

    Args:
        path (str): Dataset path on Hugging Face.
        name (Optional[str]): Name of a specific subset if any. Defaults to `None`.
    """
    return load_dataset(path, name, **config_kwargs)


def load_zenodo_dataset(
    path: str,
    entity_lexicon_or_index: dict,
    negative_type: str = "random",
    example_type: str = "triplet",
):
    """Load a HiT dataset from a local version downloaded from Zenodo.

    It is recommended to use `load_hf_dataset` from this library or `load_dataset` from HuggingFace datasets if one doesn't require the original entity IDs.

    See available datasets on: https://doi.org/10.5281/zenodo.10511042

    Args:
        path (str): Path to a local dataset downloaded from Zenodo.
        entity_lexicon_or_index (dict): A dictionary to transform entity IDs to names required by langauge models or indices (one-hot encoding) required by the static hierarchy models.
        negative_type (str): Type of negative examples. Options are `['random', 'hard']`.
        example_type (str): Type of example structure. Options are `['triplet', 'pair', 'idx']`.
    """

    assert negative_type in ["random", "hard"], f"Unknown negative type '{negative_type}'."
    assert example_type in ["triplet", "pair", "idx"], f"Unknown example type '{example_type}'."
    assert entity_lexicon_or_index is not None, "The entity transformation dictionary is not found."

    # check if train, val, test splits are all there
    datafiles = dict()
    for split in ["train", "val", "test"]:
        split_path = os.path.join(path, f"{split}.jsonl")
        if os.path.isfile(split_path):
            datafiles[split] = split_path
        else:
            logger.info(f"No {split} split available.")

    # load the jsonl dataset altogther
    dataset = load_dataset("json", data_files=datafiles)

    transform = {
        "triplet": zenodo_example_to_triplets,
        "pair": zenodo_example_to_pairs,
        "idx": zenodo_example_to_idxs,
    }[example_type]

    for split, examples in dataset.items():

        if example_type == "idx":
            # for static embedding model, inputs are not flattened
            dataset_split = [
                transform(example, negative_type, entity_lexicon_or_index)
                for example in tqdm(examples, desc=f"Map ({split})", leave=True)
            ]
        else:
            # for other models, inputs are flattened
            dataset_split = [
                transformed
                for example in tqdm(examples, desc=f"Map ({split})", leave=True)
                for transformed in transform(example, negative_type, entity_lexicon_or_index)
            ]
            dataset_split = Dataset.from_list(dataset_split)

        dataset[split] = dataset_split

    return dataset


def zenodo_example_to_triplets(example: dict, negative_type: str, entity_lexicon: dict):
    """Helper function to present Zenodo dataset examples into triplets of the form `(child, parent, negative)`."""

    child = entity_lexicon[example["child"]]["name"]
    parent = entity_lexicon[example["parent"]]["name"]
    negative_type = f"{negative_type}_negatives"
    negative_parents = [entity_lexicon[neg]["name"] for neg in example[negative_type]]
    return [{"child": child, "parent": parent, "negative": neg} for neg in negative_parents]


def zenodo_example_to_pairs(example: dict, negative_type: str, entity_lexicon: dict):
    """Helper function to present Zenodo dataset examples into labelled pairs of the form `(child, parent, label)`."""

    child = entity_lexicon[example["child"]]["name"]
    parent = entity_lexicon[example["parent"]]["name"]
    negative_type = f"{negative_type}_negatives"
    negative_parents = [entity_lexicon[neg]["name"] for neg in example[negative_type]]
    return [{"child": child, "parent": parent, "label": 1}] + [
        {"child": child, "parent": neg, "label": 0} for neg in negative_parents
    ]


def zenodo_example_to_idxs(example: dict, negative_type: str, entity_to_indices: dict):
    """Helper function to present Zenodo dataset examples into an entity index list of `(child_idx, paren_idx, *negative_idxs)`."""

    child = entity_to_indices[example["child"]]
    parent = entity_to_indices[example["parent"]]
    negative_type = f"{negative_type}_negatives"
    negative_parents = [entity_to_indices[neg] for neg in example[negative_type]]
    return [child, parent] + negative_parents



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import os.path
import warnings

import pandas as pd
import torch
from torch.utils.data import DataLoader

from hierarchy_transformers.losses import HyperbolicEntailmentConeStaticLoss, PoincareEmbeddingStaticLoss
from hierarchy_transformers.models import PoincareStaticEmbedding

from .metrics import evaluate_by_threshold, grid_search

logger = logging.getLogger(__name__)


class PoincareStaticEmbeddingEvaluator:
    """Evaluating hyperbolic static embedding models ([1] and [2]) for predicting entity hierarchical relationships.

        - [1] Poincaré Embedding by [Nickel et al., NeurIPS 2017](https://arxiv.org/abs/1705.08039).
        - [2] Hyperbolic Entailment Cone by [Ganea et al., ICML 2018](https://arxiv.org/abs/1804.01882).

    both of which lie in a unit Poincaré ball. According to [2], it is better to apply the entailment cone loss in the post-training phase of a Poincaré embedding model in [1].

    The main evaluation metrics are `Precision`, `Recall`, and `F-score`, with overall accuracy (`ACC`) and accuracy on negatives (`ACC-`) additionally reported. The results are written in a `.csv`. If a result file already exists, then values are appended.

    The labels need to be `0` for unrelated pairs and `1` for related pairs.

    Args:
        examples (list[int]): List of input examples containing entity IDs. Each example is formatted as `[child_id, parent_id, *negative_ids]`.
        batch_size (int): Evaluation batch size.
        truth_label (int, optional): Specify which label represents the truth. Defaults to `1`.
    """

    def __init__(self, eval_examples: list, batch_size: int, truth_label: int = 1):
        self.examples = eval_examples
        self.batch_size = batch_size
        self.truth_label = truth_label
        # result file
        self.results = pd.DataFrame(
            columns=["threshold", "precision", "recall", "f1", "accuracy", "accuracy_on_negatives"]
        )

    def inference(
        self,
        model: PoincareStaticEmbedding,
        loss: PoincareEmbeddingStaticLoss | HyperbolicEntailmentConeStaticLoss,
        device: torch.device,
    ):
        """The probing method of the pre-trained hyperbolic static embedding models. It output scores that indicate hierarchical relationships between entities."""

        # set up scoring function according to input loss
        # NOTE: both scores are smaller the better
        if isinstance(loss, PoincareEmbeddingStaticLoss):
            # distance scoring from [Nickel et al., NeurIPS 2017]
            def score_func(subject: torch.Tensor, objects: torch.Tensor, norm_score_weight: float = 1000.0):
                dists = loss.manifold.dist(subject, objects)
                subject_norms = subject.norm(dim=-1)
                objects_norms = objects.norm(dim=-1)
                return (1 + norm_score_weight * (objects_norms - subject_norms)) * dists

        elif isinstance(loss, HyperbolicEntailmentConeStaticLoss):
            # hyperbolic entailment cone scoring from [Ganea et al., ICML 2018]
            score_func = lambda subject, objects: loss.energy(objects, subject)
        else:
            raise ValueError(f"Unknown loss function type: {type(loss)}.")

        # set model to eval mode
        model.eval()

        # make predictions
        dataloader = DataLoader(torch.tensor(self.examples).to(device), shuffle=False, batch_size=self.batch_size)
        num_negatives = len(self.examples[0]) - 2  # each example is formatted as [child, parent, *negatives]
        scores = []
        labels = []
        with torch.no_grad():
            for batch in dataloader:
                subject, objects = model(batch)
                cur_scores = score_func(subject, objects)
                scores.append(cur_scores.reshape((-1,)))
        scores = torch.concat(scores, dim=0)
        labels = torch.tensor(
            ([self.truth_label] + [1 - self.truth_label] * num_negatives) * (int(len(scores) / (1 + num_negatives)))
        ).to(scores.device)
        assert len(labels) == len(scores)

        return scores, labels

    def __call__(
        self,
        model: PoincareStaticEmbedding,
        loss: PoincareEmbeddingStaticLoss | HyperbolicEntailmentConeStaticLoss,
        device: torch.device,
        output_path: str | None = None,
        epoch: int = -1,
        steps: int = -1,
        best_threshold: float | None = None,
    ):
        """Compute the evaluation metrics for the given model.

        Args:
            model (HierarchyTransformer): The model to evaluate.
            loss (Union[PoincareEmbeddingStaticLoss, HyperbolicEntailmentConeStaticLoss]): The training loss function decides which scoring function to be used.
            device (torch.device): The torch device used for evaluation.
            output_path (str, optional): Path to save the evaluation results `.csv` file. Defaults to `None`.
            epoch (int, optional): The epoch number. Defaults to `-1`.
            steps (int, optional): The number of steps. Defaults to `-1`.
            best_threshold (float, optional): The best overall threshold searched on a validation set (used for testing). Defaults to `None`.

        Returns:
            Dict[str, float]: A dictionary containing the evaluation metrics.
        """

        if best_threshold:
            # Testing with pre-defined hyperparameters
            logger.info(f"Evaluate on given hyperparemeters `best_threshold={best_threshold}`.")

            # Compute the scores
            scores, labels = self.inference(model=model, loss=loss, device=device)

            # Compute the evaluation metrics
            best_results = evaluate_by_threshold(
                scores=scores,
                labels=labels,
                threshold=best_threshold,
                truth_label=self.truth_label,
                smaller_scores_better=True,
            )

            # log the results
            if os.path.exists(os.path.join(output_path, "results.tsv")):
                self.results = pd.read_csv(os.path.join(output_path, "results.tsv"), sep="\t", index_col=0)
            else:
                warnings.warn("No previous `results.tsv` detected.")
            self.results.loc["testing"] = best_results
        else:
            # Validation with no pre-defined hyerparameters
            logger.info("Evaluate with grid search on hyperparameters `best_threshold` (overall threshold)")
            best_f1 = -1.0
            best_results = None

            # Compute the scores
            scores, labels = self.inference(model=model, loss=loss, device=device)

            # Perform grid search on hyperparameters
            cur_best_results = grid_search(
                scores=scores,
                labels=labels,
                threshold_granularity=1 if isinstance(loss, PoincareEmbeddingStaticLoss) else 100,
                truth_label=self.truth_label,
                smaller_scores_better=True,
                primary_metric="f1",
                best_primary_metric_value=best_f1,
                preformatted_best_results={},
            )
            if cur_best_results:
                best_results = cur_best_results
                best_f1 = best_results["f1"]

            idx = f"epoch={epoch}" if epoch != "validation" else epoch
            self.results.loc[idx] = best_results

        self.results.to_csv(os.path.join(output_path, "results.tsv"), sep="\t")

        logger.info(f"Eval results: {best_results}")

        return best_results



# Copyright 2024 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest
import torch
from geoopt.manifolds import PoincareBall

from hierarchy_transformers.models.hierarchy_transformer.hyperbolic import (
    project_onto_subspace,
    reflect_about_subspace,
)


@pytest.fixture
def manifold():
    # Create a PoincareBall manifold with a sample curvature
    return PoincareBall(c=1.0)


@pytest.fixture
def sample_point():
    # Create a sample point on the manifold
    return torch.tensor([0.3, 0.3], dtype=torch.float32)


@pytest.fixture
def normal_vector():
    # Create a sample normal vector for the subspace
    return torch.tensor([0.0, 1.0], dtype=torch.float32)


def test_reflect_about_subspace(sample_point, normal_vector):
    # Test the reflect_about_subspace function
    reflection = reflect_about_subspace(sample_point, normal_vector)

    # Check that the reflection is a torch.Tensor
    assert isinstance(reflection, torch.Tensor), "Reflection should be a torch.Tensor"

    # Check the shape of the reflection
    assert reflection.shape == sample_point.shape, "Reflection should have the same shape as the input point"

    # Verify the reflection calculation
    expected_reflection = torch.tensor([0.3, -0.3], dtype=torch.float32)
    assert torch.allclose(reflection, expected_reflection, atol=1e-6), "Reflection values do not match expected values"

    # Edge case: Normal vector cannot be zero
    with pytest.raises(ValueError):
        reflect_about_subspace(sample_point, torch.zeros_like(normal_vector))


def test_project_onto_subspace(manifold, sample_point, normal_vector):
    # Test the project_onto_subspace function
    projection = project_onto_subspace(manifold, sample_point, normal_vector)

    # Check that the projection is a torch.Tensor
    assert isinstance(projection, torch.Tensor), "Projection should be a torch.Tensor"

    # Check the shape of the projection
    assert projection.shape == sample_point.shape, "Projection should have the same shape as the input point"

    # Check that the projection lies within the Poincare Ball (norm less than 1)
    norm = torch.norm(projection)
    assert norm < 1.0, "Projected point should lie within the Poincare Ball"

    # Verify the projection calculation
    # Note: You may need to adjust this expected value based on the specifics of the PoincareBall manifold projection
    expected_projection = torch.tensor([0.27321523, 0.0000], dtype=torch.float32)
    assert torch.allclose(projection, expected_projection, atol=1e-6), "Projection values do not match expected values"



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
from collections.abc import Iterable

import torch
from geoopt.manifolds import PoincareBall
from sentence_transformers import SentenceTransformer
from sentence_transformers.models import Pooling, Transformer

from .hyperbolic import get_circum_poincareball

logger = logging.getLogger(__name__)


class HierarchyTransformer(SentenceTransformer):
    r"""Class for Hierarchy Transformer encoder (HiT), extending from [`SentenceTransformer`](https://www.sbert.net/).

    Attributes:
        embed_dim (int): The embedding dimension of this model.
        manifold (geoopt.manifolds.PoincareBall): The hyperbolic manifold (Poincaré Ball) of this model.
    """

    def __init__(
        self,
        model_name_or_path: str | None = None,
        modules: Iterable[torch.nn.Module] | None = None,
        device: str | None = None,
        revision: str | None = None,
    ):
        super().__init__(model_name_or_path=model_name_or_path, modules=modules, device=device, revision=revision)
        # PoincareBall in geoopt will be wrongly classified as a sub-module
        # so we use a dictionary to store the manifold
        self._register_buffer = {"manifold": self.get_circum_poincareball(self.embed_dim)}

    @property
    def embed_dim(self):
        return self._first_module().get_word_embedding_dimension()

    @property
    def manifold(self):
        return self._register_buffer["manifold"]

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        revision: str | None = None,
        pooling_mode: str | None = "mean",
        device: torch.device | None = None,
    ):
        """Load a pretrained model from HuggingFace hub or local repository."""
        try:
            # Load from sentence_transformers library
            pretrained_model = SentenceTransformer(model_name_or_path, device=device, revision=revision)
            transformer = pretrained_model._modules["0"]
            pooling = pretrained_model._modules["1"]
            assert isinstance(pooling, Pooling)
            logger.info(
                f"Load `{model_name_or_path}` from `sentence-transformers` with existing pooling (discard the normalising layer if any)."
            )
        except Exception:
            # Load from huggingface transformers library
            transformer = Transformer(model_name_or_path, max_seq_length=256, model_args={"revision": revision})
            pooling = Pooling(
                word_embedding_dimension=transformer.get_word_embedding_dimension(), pooling_mode=pooling_mode
            )
            logger.info(f"Load `{model_name_or_path}` from `huggingface-transformers` with '{pooling_mode}' pooling.")

        return cls(modules=[transformer, pooling], device=device)

    @staticmethod
    def get_circum_poincareball(embed_dim: int) -> PoincareBall:
        """Get a Poincaré Ball with a curvature adapted to a given embedding dimension so that it circumscribes the output embedding space of pre-trained language models."""
        manifold = get_circum_poincareball(embed_dim)
        logging.info(f"Poincare ball curvature: {manifold.c}")
        return manifold



# Copyright 2024 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os

import pytest
import torch

from hierarchy_transformers.models import HierarchyTransformer


@pytest.fixture(params=os.getenv("MODEL_PATHS", "").split(","))
def model_path(request):
    # Ensure there are valid model names
    if not request.param:
        pytest.fail("No valid model names found in the MODEL_PATHS environment variable")
    return request.param.strip()  # Strip any extra spaces


def test_hierarchy_transformer_loading(model_path):
    try:
        # Attempt to load the HierarchyTransformer model
        model = HierarchyTransformer.from_pretrained(model_path)
    except Exception as e:
        pytest.fail(f"Model failed to load: {str(e)}")

    # Check that the model is not None
    assert model is not None, "Loaded model is None"
    # Check that the model has a valid manifold attribute
    assert hasattr(model, "manifold"), "Model does not have a 'manifold' attribute"
    # Check that the manifold is an instance of PoincareBall
    from geoopt.manifolds import PoincareBall

    assert isinstance(model.manifold, PoincareBall), "Manifold is not an instance of PoincareBall"

    # Perform a basic check on the embedding dimension
    assert model.embed_dim > 0, "Embedding dimension should be greater than zero"

    # Test that the model can perform a simple forward pass
    sample_input = ["computer", "personal computer", "fruit", "berry"]
    try:
        output = model.encode(sample_input, convert_to_tensor=True)
    except Exception as e:
        pytest.fail(f"Model failed to encode input: {str(e)}")

    # Check that the output is a tensor with the expected shape
    assert isinstance(output, torch.Tensor), "Output is not a tensor"
    assert output.shape[0] == len(sample_input), "Output shape does not match the input batch size"
    assert output.shape[1] == model.embed_dim, "Output embedding dimension does not match model's embed_dim"



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import os.path
import warnings

import pandas as pd
import torch
from sentence_transformers.evaluation import SentenceEvaluator

from hierarchy_transformers import HierarchyTransformer

from .metrics import evaluate_by_threshold, grid_search

logger = logging.getLogger(__name__)


class HierarchyTransformerEvaluator(SentenceEvaluator):
    """Evaluating HiT models for predicting entity hierarchical relationships.

    The main evaluation metrics are `Precision`, `Recall`, and `F-score`, with overall accuracy (`ACC`) and accuracy on negatives (`ACC-`) additionally reported. The results are written in a `.csv`. If a result file already exists, then values are appended.

    The labels need to be `0` for unrelated pairs and `1` for related pairs.

    Args:
        child_entities (list[str]): List of child entity names.
        parent_entities (list[str]): List of parent entity names.
        labels (list[int]): List of reference labels.
        batch_size (int): Evaluation batch size.
        truth_label (int, optional): Specify which label represents the truth. Defaults to `1`.
    """

    def __init__(
        self,
        child_entities: list[str],
        parent_entities: list[str],
        labels: list[int],
        batch_size: int,
        truth_label: int = 1,
    ):
        super().__init__()
        # set primary metric for model selection
        self.primary_metric = "f1"
        # input evaluation examples
        self.child_entities = child_entities
        self.parent_entities = parent_entities
        self.labels = labels
        # eval batch size
        self.batch_size = batch_size
        # truth reference label
        self.truth_label = truth_label
        # result file
        self.results = pd.DataFrame(
            columns=["centri_weight", "threshold", "precision", "recall", "f1", "accuracy", "accuracy_on_negatives"]
        )
        # NOTE: static transformation staticmethod to do

    def inference(
        self,
        model: HierarchyTransformer,
        centri_weight: float,
        child_embeds: torch.Tensor | None = None,
        parent_embeds: torch.Tensor | None = None,
    ):
        """The default probing method of the HiT model. It output scores that indicate hierarchical relationships between entities.

        Optional `child_embeds` and `parent_embeds` are used to save time from repetitive encoding.
        """
        if child_embeds is None:
            logger.info("Encode child entities.")
            child_embeds = model.encode(
                sentences=self.child_entities, batch_size=self.batch_size, convert_to_tensor=True
            )
        if parent_embeds is None:
            logger.info("Encode parent entities.")
            parent_embeds = model.encode(
                sentences=self.parent_entities, batch_size=self.batch_size, convert_to_tensor=True
            )
        dists = model.manifold.dist(child_embeds, parent_embeds)
        child_norms = model.manifold.dist0(child_embeds)
        parent_norms = model.manifold.dist0(parent_embeds)
        return -(dists + centri_weight * (parent_norms - child_norms))

    def __call__(
        self,
        model: HierarchyTransformer,
        output_path: str | None = None,
        epoch: int = -1,
        steps: int = -1,
        best_centri_weight: float | None = None,
        best_threshold: float | None = None,
    ) -> dict[str, float]:
        """Compute the evaluation metrics for the given model.

        Args:
            model (HierarchyTransformer): The model to evaluate.
            output_path (str, optional): Path to save the evaluation results `.csv` file. Defaults to `None`.
            epoch (int, optional): The epoch number. Defaults to `-1`.
            steps (int, optional): The number of steps. Defaults to `-1`.
            best_centri_weight (float, optional): The best centripetal score weight searched on a validation set (used for testing). Defaults to `None`.
            best_threshold (float, optional): The best overall threshold searched on a validation set (used for testing). Defaults to `None`.

        Returns:
            Dict[str, float]: A dictionary containing the evaluation metrics.
        """

        # best thresholds and metric searched on validation sets
        assert (
            type(best_centri_weight) is type(best_threshold)
        ), "Inconsistent types of hyperparameters 'best_centri_weight' (centripetal score weight) and 'best_threshold' (overall threshold)"

        logger.info("Encode child entities.")
        child_embeds = model.encode(sentences=self.child_entities, batch_size=self.batch_size, convert_to_tensor=True)
        logger.info("Encode parent entities.")
        parent_embeds = model.encode(
            sentences=self.parent_entities, batch_size=self.batch_size, convert_to_tensor=True
        )

        if best_centri_weight and best_threshold:
            # Testing with pre-defined hyperparameters
            logger.info(
                f"Evaluate on given hyperparemeters `best_centri_weight={best_centri_weight}` (centripetal score weight) and `best_threshold={best_threshold}` (overall threshold)."
            )

            # Compute the scores
            scores = self.inference(
                model=model,
                centri_weight=best_centri_weight,
                child_embeds=child_embeds,
                parent_embeds=parent_embeds,
            )

            # Compute the evaluation metrics
            best_results = {"centri_weight": best_centri_weight}
            best_results.update(
                evaluate_by_threshold(
                    scores=scores,
                    labels=torch.tensor(self.labels).to(scores.device),
                    threshold=best_threshold,
                    truth_label=self.truth_label,
                    smaller_scores_better=False,
                )
            )

            # log the results
            if os.path.exists(os.path.join(output_path, "results.tsv")):
                self.results = pd.read_csv(os.path.join(output_path, "results.tsv"), sep="\t", index_col=0)
            else:
                warnings.warn("No previous `results.tsv` detected.")
            self.results.loc["testing"] = best_results
        else:
            # Validation with no pre-defined hyerparameters
            logger.info(
                "Evaluate with grid search on hyperparameters `best_centri_weight` (centripetal score weight) and `best_threshold` (overall threshold)."
            )
            best_f1 = -1.0
            best_results = None
            is_updated = True

            for centri_weight in range(50):
                # early stop if increasing the centri score weight does not help
                if not is_updated:
                    break
                is_updated = False

                centri_weight /= 10

                # Compute the scores
                scores = self.inference(
                    model=model, centri_weight=centri_weight, child_embeds=child_embeds, parent_embeds=parent_embeds
                )

                # Perform grid search on hyperparameters
                cur_best_results = grid_search(
                    scores=scores,
                    labels=torch.tensor(self.labels).to(scores.device),
                    threshold_granularity=100,
                    truth_label=self.truth_label,
                    smaller_scores_better=False,
                    primary_metric="f1",
                    best_primary_metric_value=best_f1,
                    preformatted_best_results={"centri_weight": centri_weight},
                )
                if cur_best_results:
                    best_results = cur_best_results
                    best_f1 = best_results["f1"]
                    is_updated = True

            idx = f"epoch={epoch}" if epoch != "validation" else epoch
            self.results.loc[idx] = best_results

        self.results.to_csv(os.path.join(output_path, "results.tsv"), sep="\t")

        logger.info(f"Eval results: {best_results}")

        return best_results



# Copyright 2023 Yuan He

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import os.path
import warnings
from string import Template

import pandas as pd
import torch
from sentence_transformers import SentenceTransformer
from sentence_transformers.evaluation import SentenceEvaluator

from .metrics import evaluate_by_threshold, grid_search

logger = logging.getLogger(__name__)


class SentenceTransformerEvaluator(SentenceEvaluator):
    """Evaluating sBERT models for predicting entity hierarchical relationships.

    The main evaluation metrics are `Precision`, `Recall`, and `F-score`, with overall accuracy (`ACC`) and accuracy on negatives (`ACC-`) additionally reported. The results are written in a `.csv`. If a result file already exists, then values are appended.

    The labels need to be `0` for unrelated pairs and `1` for related pairs.

    Args:
        child_entities (list[str]): List of child entity names.
        parent_entities (list[str]): List of parent entity names.
        labels (list[int]): List of reference labels.
        batch_size (int): Evaluation batch size.
        truth_label (int, optional): Specify which label represents the truth. Defaults to `1`.
        template (str, optional): The probing template.
    """

    def __init__(
        self,
        child_entities: list[str],
        parent_entities: list[str],
        labels: list[int],
        batch_size: int,
        truth_label: int = 1,
        template: str = "${child} is a ${parent}.",
    ):
        super().__init__()
        # set primary metric for model selection
        self.primary_metric = "f1"
        # input evaluation examples
        self.child_entities = child_entities
        self.parent_entities = parent_entities
        self.labels = labels
        # eval batch size
        self.batch_size = batch_size
        # truth reference label
        self.truth_label = truth_label
        # result file
        self.results = pd.DataFrame(
            columns=["threshold", "precision", "recall", "f1", "accuracy", "accuracy_on_negatives"]
        )
        # template for probing
        self.template = Template(template)

    def inference(self, model: SentenceTransformer):
        """The probing method of the pre-trained sBERT model. It output scores that indicate hierarchical relationships between entities."""
        sentences = []
        masked_sentences = []
        for child, parent in zip(self.child_entities, self.parent_entities):
            sentences.append(self.template.substitute(child=child, parent=parent))
            masked_sentences.append(self.template.substitute(child=child, parent=model.tokenizer.mask_token))

        sentence_embeds = model.encode(sentences=sentences, convert_to_tensor=True, show_progress_bar=True)
        masked_embeds = model.encode(sentences=masked_sentences, convert_to_tensor=True, show_progress_bar=True)

        # use the cosine similarity between masked and
        return torch.cosine_similarity(masked_embeds, sentence_embeds)

    def __call__(
        self,
        model: SentenceTransformer,
        output_path: str | None = None,
        epoch: int = -1,
        steps: int = -1,
        best_threshold: float | None = None,
    ) -> dict[str, float]:
        """Compute the evaluation metrics for the given model.

        Args:
            model (HierarchyTransformer): The model to evaluate.
            output_path (str, optional): Path to save the evaluation results `.csv` file. Defaults to `None`.
            epoch (int, optional): The epoch number. Defaults to `-1`.
            steps (int, optional): The number of steps. Defaults to `-1`.
            best_centri_weight (float, optional): The best centripetal score weight searched on a validation set (used for testing). Defaults to `None`.
            best_threshold (float, optional): The best overall threshold searched on a validation set (used for testing). Defaults to `None`.

        Returns:
            Dict[str, float]: A dictionary containing the evaluation metrics.
        """

        if best_threshold:
            # Testing with pre-defined hyperparameters
            logger.info(f"Evaluate on given hyperparemeters `best_threshold={best_threshold}`.")

            # Compute the scores
            scores = self.inference(model=model)

            # Compute the evaluation metrics
            best_results = evaluate_by_threshold(
                scores=scores,
                labels=torch.tensor(self.labels).to(scores.device),
                threshold=best_threshold,
                truth_label=self.truth_label,
                smaller_scores_better=False,
            )

            # log the results
            if os.path.exists(os.path.join(output_path, "results.tsv")):
                self.results = pd.read_csv(os.path.join(output_path, "results.tsv"), sep="\t", index_col=0)
            else:
                warnings.warn("No previous `results.tsv` detected.")
            self.results.loc["testing"] = best_results
        else:
            # Validation with no pre-defined hyerparameters
            logger.info("Evaluate with grid search on hyperparameters `best_threshold` (overall threshold)")
            best_f1 = -1.0
            best_results = None

            # Compute the scores
            scores = self.inference(model=model)

            # Perform grid search on hyperparameters
            cur_best_results = grid_search(
                scores=scores,
                labels=torch.tensor(self.labels).to(scores.device),
                threshold_granularity=1000,
                truth_label=self.truth_label,
                smaller_scores_better=False,
                primary_metric="f1",
                best_primary_metric_value=best_f1,
                preformatted_best_results={},
            )
            if cur_best_results:
                best_results = cur_best_results
                best_f1 = best_results["f1"]

            idx = f"epoch={epoch}" if epoch != "validation" else epoch
            self.results.loc[idx] = best_results

        self.results.to_csv(os.path.join(output_path, "results.tsv"), sep="\t")

        logger.info(f"Eval results: {best_results}")

        return best_results
